# Referenced https://github.com/GaiZhenbiao/ChuanhuChatGPT project

import importlib
import json
import traceback

import requests
from loguru import logger
from .config import load_config

timeout_bot_msg = (
    "[local] Request timeout, network error. please check proxy settings in config.py."
)


CONFIGS = load_config()


def get_full_error(chunk, stream_response):
    """Get the complete error message returned from Openai."""
    while True:
        try:
            chunk += next(stream_response)
        except StopIteration:
            break
    return chunk


def predict_no_ui(inputs, top_p, temperature, history=[]):
    """Send to chatGPT, wait for reply, complete in one go, and do not display intermediate process.

    A simplified version of the predict function.
    Used for cases where the payload is relatively large, or for implementing complex functions with multiple threads and nesting.

    inputs is the input for this inquiry
    top_p, temperature are internal tuning parameters of chatGPT
    history is the previous conversation list
    (Note that if either inputs or history is too long, it will trigger a token count overflow error, and then raise ConnectionAbortedError.)
    """
    headers, payload = generate_payload(
        inputs, top_p, temperature, history, system_prompt="", stream=False
    )

    retry = 0
    while True:
        try:
            # make a POST request to the API endpoint, stream=False
            response = requests.post(
                CONFIGS.API_URL,
                headers=headers,
                proxies=CONFIGS.proxies,
                json=payload,
                stream=False,
                timeout=CONFIGS.TIMEOUT_SECONDS * 2,
            )
            break

        except requests.exceptions.ReadTimeout:
            retry += 1
            traceback.print_exc()
            if CONFIGS.MAX_RETRY != 0:
                logger.info(
                    f"Request timed out, retrying ({retry}/{CONFIGS.MAX_RETRY}) ……"
                )

            if retry > CONFIGS.MAX_RETRY:
                raise TimeoutError

    try:
        result = json.loads(response.text)["choices"][0]["message"]["content"]
        return result
    except Exception:
        if "choices" not in response.text:
            logger.info(response.text)
        raise ConnectionAbortedError("Json parsing is not normal" + response.text)


def predict(
    inputs,
    top_p,
    temperature,
    chatbot=[],
    history=[],
    system_prompt="",
    stream=True,
    additional_fn=None,
):
    """Send to chatGPT, and get output in a streaming manner.
    Used for basic conversation functions.
    inputs is the input for this inquiry
    top_p, temperature are internal tuning parameters of chatGPT
    history is the previous conversation list (Note that if either inputs or history is too long, it will trigger a token count overflow error.)
    chatbot is the conversation list displayed in the WebUI. Modify it, then yeild it out, and you can directly modify the content of the conversation interface.
    additional_fn represents which button was clicked. The buttons are in functional.py.
    """
    if additional_fn is not None:
        from . import functional

        importlib.reload(functional)
        functional = functional.get_functionals()
        inputs = (
            functional[additional_fn]["Prefix"]
            + inputs
            + functional[additional_fn]["Suffix"]
        )

    if stream:
        raw_input = inputs
        logger.info(f"[raw_input] {raw_input}")
        chatbot.append((inputs, ""))
        yield chatbot, history, "Waiting for response"

    headers, payload = generate_payload(
        inputs, top_p, temperature, history, system_prompt, stream
    )
    history.append(inputs)
    history.append(" ")

    retry = 0
    while True:
        try:
            # make a POST request to the API endpoint, stream=True
            response = requests.post(
                CONFIGS.API_URL,
                headers=headers,
                proxies=CONFIGS.proxies,
                json=payload,
                stream=True,
                timeout=CONFIGS.TIMEOUT_SECONDS,
            )
            break
        except Exception:
            retry += 1
            chatbot[-1] = (chatbot[-1][0], timeout_bot_msg)
            retry_msg = (
                f", retrying ({retry}/{CONFIGS.MAX_RETRY}) ……"
                if CONFIGS.MAX_RETRY > 0
                else ""
            )
            yield chatbot, history, "Request timed out" + retry_msg
            if retry > CONFIGS.MAX_RETRY:
                raise TimeoutError

    gpt_replying_buffer = ""

    is_head_of_the_stream = True
    if stream:
        stream_response = response.iter_lines()
        while True:
            chunk = next(stream_response)
            if is_head_of_the_stream:
                # The first frame of the data stream does not carry content
                is_head_of_the_stream = False
                continue

            if chunk:
                try:
                    if len(json.loads(chunk.decode()[6:])["choices"][0]["delta"]) == 0:
                        # Determine the end of the data stream, and gpt_replying_buffer is also written
                        logger.info(f"[response] {gpt_replying_buffer}")
                        break
                    # Process the body of the data stream
                    chunkjson = json.loads(chunk.decode()[6:])
                    status_text = (
                        f"finish_reason: {chunkjson['choices'][0]['finish_reason']}"
                    )
                    # If an exception is thrown here, it is generally because the text is too long. See the output of get_full_error for details.
                    gpt_replying_buffer = (
                        gpt_replying_buffer
                        + json.loads(chunk.decode()[6:])["choices"][0]["delta"][
                            "content"
                        ]
                    )
                    history[-1] = gpt_replying_buffer
                    chatbot[-1] = (history[-2], history[-1])
                    yield chatbot, history, status_text

                except Exception:
                    traceback.print_exc()
                    yield chatbot, history, "Json parsing is not normal."

                    chunk = get_full_error(chunk, stream_response)
                    error_msg = chunk.decode()
                    if "reduce the length" in error_msg:
                        chatbot[-1] = (
                            chatbot[-1][0],
                            "[Local Message] Input (or history) is too long, please reduce input or clear history by refleshing this page.",
                        )
                        history = []
                    elif "Incorrect API key" in error_msg:
                        chatbot[-1] = (
                            chatbot[-1][0],
                            "[Local Message] Incorrect API key provided.",
                        )

                    else:
                        from .utils import regular_txt_to_markdown

                        tb_str = regular_txt_to_markdown(traceback.format_exc())
                        chatbot[-1] = (
                            chatbot[-1][0],
                            f"[Local Message] Json Error \n\n {tb_str} \n\n {regular_txt_to_markdown(chunk.decode()[4:])}",
                        )

                    yield chatbot, history, "Json parsing is not normal" + error_msg
                    return


def generate_payload(inputs, top_p, temperature, history, system_prompt, stream):
    """Integrate all information, select LLM model, generate http request, and prepare for sending requests."""
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {CONFIGS.API_KEY}",
    }

    conversation_cnt = len(history) // 2

    messages = [{"role": "system", "content": system_prompt}]
    if conversation_cnt:
        for index in range(0, 2 * conversation_cnt, 2):
            what_i_have_asked = {}
            what_i_have_asked["role"] = "user"
            what_i_have_asked["content"] = history[index]
            what_gpt_answer = {}
            what_gpt_answer["role"] = "assistant"
            what_gpt_answer["content"] = history[index + 1]
            if what_i_have_asked["content"] != "":
                if what_gpt_answer["content"] == "":
                    continue
                if what_gpt_answer["content"] == timeout_bot_msg:
                    continue
                messages.append(what_i_have_asked)
                messages.append(what_gpt_answer)
            else:
                messages[-1]["content"] = what_gpt_answer["content"]

    what_i_ask_now = {}
    what_i_ask_now["role"] = "user"
    what_i_ask_now["content"] = inputs
    messages.append(what_i_ask_now)

    payload = {
        "model": CONFIGS.LLM_MODEL,
        "messages": messages,
        "temperature": temperature,  # 1.0,
        "top_p": top_p,  # 1.0,
        "n": 1,
        "stream": stream,
        "presence_penalty": 0,
        "frequency_penalty": 0,
    }

    logger.info(f" {CONFIGS.LLM_MODEL} : {conversation_cnt} : {inputs}")
    return headers, payload



import toml
from loguru import logger
from pydantic import BaseModel
from pathlib import Path
from typing import Optional


class Config(BaseModel):
    API_KEY: str
    API_URL: str = "https://api.openai.com/v1/chat/completions"
    USE_PROXY: bool = False
    # After sending the request to OpenAI, how long to wait before it times out
    TIMEOUT_SECONDS: int = 20
    # The port of the webpage, -1 means random port
    WEB_PORT: int = -1
    # If OpenAI does not respond (network congestion, proxy failure, KEY invalid), the number of retries is limited
    MAX_RETRY: int = 2
    LLM_MODEL: str = "gpt-3.5-turbo"
    proxies: Optional[dict] = None


CACHE_CONFIGS = None


def read_config():
    config_file = Path("chataca.toml")

    if not config_file.exists():
        config_file = Path.home() / ".config" / "chataca" / "chataca.toml"

    if not config_file.exists():
        logger.error("Config file not found")
        raise FileNotFoundError("Config file not found")

    with open(config_file, "r") as f:
        config = toml.load(f)

    configs = Config(**config)

    if configs.USE_PROXY:
        configs.proxies = {
            "http": "socks5h://localhost:11284",
            "https": "socks5h://localhost:11284",
        }
        logger.info("Network proxy status: running.")
    else:
        logger.info(
            "Network proxy status: not configured. In the absence of a proxy, it is likely that you will not be able to access it."
        )

    return configs


def load_config():
    global CACHE_CONFIGS
    if CACHE_CONFIGS is None:
        CACHE_CONFIGS = read_config()
    return CACHE_CONFIGS



"""This module contains a function for checking the validity of a proxy server."""
import requests
from loguru import logger


def check_proxy(proxies):
    """Check the validity of a proxy server.

    Args:
    ----
        proxies (dict): A dictionary containing the proxy server information.

    Returns:
    -------
        None
    """

    proxies_https = None if proxies is None else proxies.get("https")

    try:
        response = requests.get("https://ipapi.co/json/", proxies=proxies, timeout=4)
        response.raise_for_status()
        data = response.json()
        country = data["country_name"]
        logger.info(f"Proxy server {proxies_https} is located in {country}")

    except requests.exceptions.RequestException as e:
        logger.error(f"An error occurred: {e}")



from loguru import logger
from ..utils import (
    catch_exception,
    predict_no_ui_but_counting_down,
    report_execption,
    write_results_to_file,
)

from pathlib import Path
import time
import os


FAST_DEBUG = False


def parse_paper(
    file_manifest, project_folder, top_p, temperature, chatbot, history, systemPromptTxt
):
    logger.info(f"begin analysis on: {file_manifest}")

    for index, fp in enumerate(file_manifest):
        with open(fp, encoding="utf-8") as f:
            file_content = f.read()

        prefix = (
            "Next, please analyze the following paper files one by one and summarize their contents"
            if index == 0
            else ""
        )
        i_say = (
            prefix
            + f"Please summarize the following article fragment in Chinese, the file name is {os.path.relpath(fp, project_folder)}, and the article content is ```{file_content}```"
        )
        i_say_show_user = (
            prefix
            + f"[{index}/{len(file_manifest)}] Please summarize the following article fragment in Chinese: {os.path.abspath(fp)}"
        )
        chatbot.append((i_say_show_user, "[Local Message] waiting gpt response."))
        logger.info("[1] yield chatbot, history")
        yield chatbot, history, "normal"

        if not FAST_DEBUG:
            msg = "normal"
            # ** gpt request **
            gpt_say = yield from predict_no_ui_but_counting_down(
                i_say, i_say_show_user, chatbot, top_p, temperature, history=[]
            )  # with timeout countdown

            logger.info("[2] end gpt req")
            chatbot[-1] = (i_say_show_user, gpt_say)
            history.append(i_say_show_user)
            history.append(gpt_say)
            logger.info("[3] yield chatbot, history")
            yield chatbot, history, msg
            logger.info("[4] next")
            if not FAST_DEBUG:
                time.sleep(2)

    all_file = ", ".join(
        [os.path.relpath(fp, project_folder) for index, fp in enumerate(file_manifest)]
    )
    i_say = f"Based on your analysis above, summarize the entire text, write a paragraph of Chinese abstract in academic language, and then write a paragraph of English abstract (including {all_file})."
    chatbot.append((i_say, "[Local Message] waiting gpt response."))
    yield chatbot, history, "normal"

    if not FAST_DEBUG:
        msg = "normal"
        # ** gpt request **
        gpt_say = yield from predict_no_ui_but_counting_down(
            i_say, i_say, chatbot, top_p, temperature, history=history
        )  # with timeout countdown

        chatbot[-1] = (i_say, gpt_say)
        history.append(i_say)
        history.append(gpt_say)
        yield chatbot, history, msg
        res = write_results_to_file(history)
        chatbot.append(("Are you done?", res))
        yield chatbot, history, msg


@catch_exception
def read_artical_write_abstract(
    txt, top_p, temperature, chatbot, history, systemPromptTxt, WEB_PORT
):
    history = []  # clear history to avoid input overflow

    project_folder = Path(txt)
    if not project_folder.exists or not project_folder.is_dir():
        msg = "empty input field or path is not valid"
        report_execption(
            chatbot,
            history,
            a=f"parse project: {msg}",
            b=f"cannot find local project or access denied: {msg}",
        )
        yield chatbot, history, "normal"

    file_manifest = list(project_folder.rglob("*.tex"))

    if len(file_manifest) == 0:
        report_execption(
            chatbot,
            history,
            a=f"Parse project: {txt}",
            b=f"Cannot find any .tex files: {txt}",
        )
        yield chatbot, history, "normal"

    yield from parse_paper(
        file_manifest,
        project_folder,
        top_p,
        temperature,
        chatbot,
        history,
        systemPromptTxt,
    )



# This program is written by: https://github.com/polarwinkel/mdtex2html

import re

from latex2mathml.converter import convert as tex2mathml

incomplete = '<font style="color:orange;" class="tooltip">&#9888;<span class="tooltiptext">formula incomplete</span></font>'
convError = '<font style="color:red" class="tooltip">&#9888;<span class="tooltiptext">LaTeX-convert-error</span></font>'


def convert(mdtex, extensions=[], splitParagraphs=True):
    """converts recursively the Markdown-LaTeX-mixture to HTML with MathML."""
    found = False
    # handle all paragraphs separately (prevents aftereffects)
    if splitParagraphs:
        parts = re.split("\n\n", mdtex)
        result = ""
        for part in parts:
            result += convert(part, extensions, splitParagraphs=False)
        return result

    # find first $$-formula:
    parts = re.split("\\${2}", mdtex, 2)

    if len(parts) > 1:
        found = True
        result = convert(parts[0], extensions, splitParagraphs=False) + "\n"
        try:
            result += '<div class="blockformula">' + tex2mathml(parts[1]) + "</div>\n"
        except Exception:
            result += '<div class="blockformula">' + convError + "</div>"
        if len(parts) == 3:
            result += convert(parts[2], extensions, splitParagraphs=False)
        else:
            result += '<div class="blockformula">' + incomplete + "</div>"
    # else find first $-formulas:
    else:
        parts = re.split("\\${1}", mdtex, 2)

    if len(parts) > 1 and not found:
        found = True
        try:
            mathml = tex2mathml(parts[1])
        except Exception:
            mathml = convError
        if (
            parts[0].endswith("\n\n") or parts[0] == ""
        ):  # make sure textblock starts before formula!
            parts[0] = parts[0] + "&#x200b;"
        if len(parts) == 3:
            result = convert(
                parts[0] + mathml + parts[2], extensions, splitParagraphs=False
            )
        else:
            result = convert(
                parts[0] + mathml + incomplete, extensions, splitParagraphs=False
            )
    # else find first \[..\]-equation:
    else:
        parts = re.split(r"\\\[", mdtex, 1)

    if len(parts) > 1 and not found:
        found = True
        result = convert(parts[0], extensions, splitParagraphs=False) + "\n"
        parts = re.split(r"\\\]", parts[1], 1)
        try:
            result += '<div class="blockformula">' + tex2mathml(parts[0]) + "</div>\n"
        except Exception:
            result += '<div class="blockformula">' + convError + "</div>"
        if len(parts) == 2:
            result += convert(parts[1], extensions, splitParagraphs=False)
        else:
            result += '<div class="blockformula">' + incomplete + "</div>"
    # else find first \(..\)-equation:
    else:
        parts = re.split(r"\\\(", mdtex, 1)

    if len(parts) > 1 and not found:
        found = True
        subp = re.split(r"\\\)", parts[1], 1)
        try:
            mathml = tex2mathml(subp[0])
        except Exception:
            mathml = convError
        if (
            parts[0].endswith("\n\n") or parts[0] == ""
        ):  # make sure textblock starts before formula!
            parts[0] = parts[0] + "&#x200b;"
        if len(subp) == 2:
            result = convert(
                parts[0] + mathml + subp[1], extensions, splitParagraphs=False
            )
        else:
            result = convert(
                parts[0] + mathml + incomplete, extensions, splitParagraphs=False
            )

    if not found:
        result = mdtex

    return result
